/*******************************************************************
cwrapper.c                                             

Author: Scott Hawley, shawley@physics.utexas.edu

This is technically the main routine for the multigrid solver, where
we read command line arguments, allocate a big block of memory, and 
then call Fortran routine robin3d, which is where the actual 
multigrid method is implemented
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>


/*******************************************************************
*******************************************************************/
double read_next_val(FILE *infile, double default_val)
{
   char s[256];
   int  valread;
   double rval;

   valread = 0;

   while (!feof(infile) && !valread) {
      fgets( s, 255, infile);
      if (s[0] != '#') {
          if (s[0] == '.') { 
             rval = default_val;
          } else {
            sscanf(s,"%lf", &rval);
          }
        valread = 1;
      }
   }

   return rval;
}


/*******************************************************************
*******************************************************************/
int parse_param_file(char *filename,
                     int *eqtype, int *exact_type, double *M,
                     double *sigma, double *Ksq, double *Asq,
                     int *maxlev, int *minlev, 
                     double *xyzmax, double *xyzmin,
                     double *r_mask, double *mdisp_x,
                     int *nvcycle, int *preswp, int *pstswp,
                     int *gft_trace) 
{
   FILE *infile;
   int  rc;
   double rval;

   infile = fopen(filename,"r");
   if (infile == NULL) {
      fprintf(stderr,"Error opening file '%s'.\n",filename);
      return -1;
   }

   *eqtype =     (int)read_next_val(infile, 2.0);
   *exact_type = (int)read_next_val(infile, 2.0);
   *M =               read_next_val(infile, 1.0);
   *sigma =           read_next_val(infile,-1.0);
   *Ksq =             read_next_val(infile, 1.0);
   *Asq =             read_next_val(infile, 1.0);
   *maxlev =     (int)read_next_val(infile, 3.0);
   *minlev =     (int)read_next_val(infile, 3.0);
   *xyzmax =          read_next_val(infile, 5.0);
   *xyzmin =          read_next_val(infile,-5.0);
   *r_mask =          read_next_val(infile, 2.0);
   *mdisp_x =         read_next_val(infile, 0.0);
   *nvcycle =    (int)read_next_val(infile, 1.0);
   *preswp =     (int)read_next_val(infile, 3.0);
   *pstswp =     (int)read_next_val(infile, 3.0);
   *gft_trace =  (int)read_next_val(infile, 0.0);
   
   fclose(infile);

   return 0;
}
                     

/*******************************************************************
showusage: prints usage message
*******************************************************************/
void showusage()
{
      fprintf(stderr,"usage: robin3d <par_file>\n See a sample par file for par file syntax\n");
      exit(1);
}


/*******************************************************************
Main routine
*******************************************************************/
int main(int argc, char **argv)
{
   int         qsize,         mallocsize;
   static      double         *q;   /* static means it's put on heap */

   int         level,         lmin;
   double      mrad,          mdisp_x;
   double      sigma,         Ksq,            Asq; 
   int         nvcycle,       preswp,         pstswp;
   int         gft_trace;

   int         eqtype,        exact_type;
   double      M;
   double      xyzmax,        xyzmin;
   int         rc;


/*----------------------------------
    Parse command-line arguments
------------------------------------*/
   if (argc < 2) {
      showusage();
   }

   rc = parse_param_file(argv[1],
                     &eqtype, &exact_type, &M,
                     &sigma, &Ksq, &Asq,
                     &level, &lmin, 
                     &xyzmax, &xyzmin,
                     &mrad,  &mdisp_x,
                     &nvcycle, &preswp, &pstswp,
                     &gft_trace);

   printf("Params = %d, %d, %f, %f, %f ,%f %d, %d, %f, %f, %f, %f %d, %d, %d, %d\n",
                     eqtype, exact_type, M,
                     sigma, Ksq, Asq,
                     level, lmin, 
                     xyzmax, xyzmin,
                     mrad,  mdisp_x,
                     nvcycle, preswp, pstswp,
                     gft_trace);


   if (level < 1) {
      showusage();
   }

   if (level < lmin) {
      fprintf(stderr,"robin3d: Error: level(=%d) < lmin(=%d)\n",level,lmin);
      exit(1);
   }

/*----------------------------------
    Allocate memory
------------------------------------*/
   qsize = 100000000;
   fprintf(stderr,"cwrapper: gft_trace =%d\n",gft_trace);
   mallocsize =  (int)qsize * (int)sizeof(double) ;
   fprintf(stderr,"qsize = %d\n",qsize);
   fprintf(stderr,"mallocsize = %d\n",mallocsize);
   q = (double *)malloc( mallocsize );

/*----------------------------------
    Call multigrid solver
------------------------------------*/
   robin3d_(q, &qsize, &level, &lmin, &mrad, &mdisp_x, &sigma,
            &nvcycle, &preswp, &pstswp, &gft_trace,
            &eqtype, &exact_type, &M, &Ksq, &Asq, &xyzmax, &xyzmin);

/*----------------------------------
    Normal exit
------------------------------------*/
  exit(0);
}
