c======================================================================
c  maskutils: Utilities for managing values inside the excision mask
c             and the outer boundary
c
c  Scott Hawley  shawley@physics.utexas.edu
c======================================================================


c======================================================================
c     minit: Initialize the mask function
c
c     This will initialize the mask function with two shapes,
c     e.g. two circles, which are on either side of the y axis
c
c
c
c  INPUTS:
c  maxlev 
c  lmin 
c  mrad     Physical Radius of mask. Negative value turns excision off.
c  mdisp_x "x-displacement of center of each mask shape".  Choosing
c              mdisp_x = 0 will give only one (cubical or spherical) 
c              excision region.  A nonzero mdisp_x will result in two 
c              identical excision regions, with centers separated by 
c              mdisp_x.  (Note that if 0 < mdisp_x < mrad,  then one 
c              will get one exicsion region (e.g., a "peanut" for 
c              spherical shapes)
c
c  Other inputs (mask arrays) are obtained via common blocks defined
c     in mg.inc
c
c  OUTPUTS
c  Values on qm(mask) are set.  qm is obtained via common block in 
c     mg.inc
c======================================================================
      subroutine minit(maxlev, lmin, mrad, mdisp_x, q, qsize)
            
      implicit none
     
      include    'mg.inc'
      include    'maskcodes.inc'
               
      integer    maxlev,           lmin
      integer    mpcount
      real*8     mrad
      real*8     mdisp_x
      integer    qsize
      real*8     q(qsize)

c Local variables
      integer    l

         
      mask(lmin) = 1
      do l = lmin, maxlev

         write(0,*)' Setting mask on level ',l

         if (l .gt. lmin) mask(l) = mask(l-1) + (2**(l-1) + 1)**3

         call imls  ( qm(mask(l)), m_int, nx(l), ny(l), nz(l) )

         write(0,*) 'Calling mset'
         call mset  ( qm(mask(l)), nx(l), ny(l), nz(l), 
     &                q(x(l)), q(y(l)), q(z(l)), mrad, 
     &                mdisp_x, mpcount, l-1 )

         if (l .le. 3) then
             call mprint( qm(mask(l)), nx(l), ny(l), nz(l),
     &          1, nx(l), 1, ny(l), nz(l)/2+1, nz(l)/2+1)
         endif

      enddo

      return
      end



c======================================================================
c  mset: This actually sets the values in the mask
c
c  INPUTS:
c  nx, ny, nz
c  x,  y,  z
c  mrad
c  mdisp_x
c  esr_size  Width (horizontal & vertical, not diagonal) of the 
c               extra smoothing region (ESR)
c
c  OUTPUTS:
c  mask
c  mpcount    The number of points which are either excised or inner
c               boundary points.  This is here for historical reasons,
c               an not an important part of the code.
c======================================================================
      subroutine  mset(mask, nx, ny, nz, x, y, z, 
     &                 mrad, mdisp_x, mpcount, esr_size)

      implicit none

      include 'maskcodes.inc'
      include 'mgprob.inc'

      integer  nx,            ny,                nz
      integer  mask(nx, ny,nz)
      real*8   x(nx),         y(ny),             z(nz)
      real*8   mrad,          mdisp_x
      integer  mpcount
      integer  esr_size

c  Local variables
      integer  i,             j,                 k
      integer  loop
      real*8   rad
      real*8   morig_x,       morig_y,           morig_z

      logical     msquare
      parameter ( msquare = .false. )

      mpcount = 0

c------------------------------------------------------------------
c     Zero out mask
c------------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               mask(i,j,k) = m_int
            end do
         end do
      end do

c------------------------------------------------------------------
c    Set up two "holes", separated by distance mdisp_x
c------------------------------------------------------------------
      morig_x = mdisp_x
      morig_y = 0.0d0
      morig_z = 0.0d0

c    set origin for mask to be offset from the center of the grid
c    (which may not be x,y,z=0, but rather x,y,z=0.5)
      if (excttype .eq. 1) then
        morig_x = morig_x + (x(nx)-x(1))/2.0d0 
        morig_y = morig_y + (y(ny)-y(1))/2.0d0 
        morig_z = morig_z + (z(nz)-z(1))/2.0d0 
      endif

      do loop = 1, 2

         if (.not. msquare) then

c  Spherical holes 
            do k = 1, nz
              do j = 1, ny
                do i = 1, nx

                   rad =  sqrt (
     &                (x(i) - morig_x)**2 +
     &                (y(j) - morig_y)**2 +
     &                (z(k) - morig_z)**2 )
                   if ( rad .le. mrad) then
                      mask(i,j,k) = m_ex
                      mpcount = mpcount + 1
                  endif

               enddo
              enddo
            enddo

         else

c  Cubical holes 
            do k = 1, nz
              do j = 1, ny
                do i = 1, nx

                   if ( abs(x(i) - morig_x) .le. mrad  .and. 
     &                  abs(y(j) - morig_y) .le. mrad  .and.
     &                  abs(z(k) - morig_z) .le. mrad) then
                     mask(i,j,k) = m_ex
                     mpcount = mpcount + 1
                   endif

                 enddo
              enddo
            enddo
c   End of cubical/spherical section
         endif


c-----------------------------------------------------------------
c   For next pass through the (# of holes) loop, use a negative 
c   origin morig_x for the center of the mask shape
c-----------------------------------------------------------------
         morig_x = morig_x - 2.0d0 * mdisp_x

      enddo

c-----------------------------------------------------------------
c    Set inner boundary as outermost part of excision mask
c-----------------------------------------------------------------
        do k = 1, nz
          do j = 1, ny
            do i = 1, nx

              if ( (mask(i,j,k) .eq. m_ex)  .and. (
     &              mask(i,j,k+1) .eq. m_int .or.
     &              mask(i,j,k-1) .eq. m_int .or.
     &              mask(i,j+1,k) .eq. m_int .or.
     &              mask(i+1,j,k) .eq. m_int .or.
     &              mask(i-1,j,k) .eq. m_int .or.
     &              mask(i,j-1,k) .eq. m_int 
 
c------------------------------------------------------
c  "diagonal" points:
c  Note on the following block of code:
c  If mrad is not large enough, and the following code
c  is enabled, it can result in the point at x=y=z=0 becoming
c  an inner boundary point rather than an excised point.
c  This will be bad since the the point at x=y=z=0 will then
c  receive a value of -Infinity.
c 
c  Generally I leave this block commented out.  This has
c  little effect, since the restriction operator (dmrshw, in
c  robin3dlib.f) is set up to "copy" if any of the nearby 
c  points is an inner boundary point OR an excised point.
c------------------------------------------------------
c     &         .or. mask(i,j+1,k+1) .eq. m_int .or.
c     &              mask(i,j+1,k-1) .eq. m_int .or.
c     &              mask(i,j-1,k+1) .eq. m_int .or.
c     &              mask(i,j-1,k-1) .eq. m_int .or.
c     &              mask(i+1,j,k+1) .eq. m_int .or.
c     &              mask(i+1,j,k-1) .eq. m_int .or.
c     &              mask(i-1,j,k+1) .eq. m_int .or.
c     &              mask(i-1,j,k-1) .eq. m_int .or.
c     &              mask(i+1,j+1,k) .eq. m_int .or.
c     &              mask(i+1,j-1,k) .eq. m_int .or.
c     &              mask(i-1,j+1,k) .eq. m_int .or.
c     &              mask(i-1,j-1,k) .eq. m_int .or.
cc
c     &              mask(i+1,j+1,k+1) .eq. m_int .or.
c     &              mask(i+1,j+1,k-1) .eq. m_int .or.
c     &              mask(i+1,j-1,k+1) .eq. m_int .or.
c     &              mask(i+1,j-1,k-1) .eq. m_int .or.
c     &              mask(i-1,j+1,k+1) .eq. m_int .or.
c     &              mask(i-1,j+1,k-1) .eq. m_int .or.
c     &              mask(i-1,j-1,k+1) .eq. m_int .or.
c     &              mask(i-1,j-1,k-1) .eq. m_int 
c------------------------------------------------------
c  End comment regarding "the following block of code"
c------------------------------------------------------
     &              ) ) then
                  mask(i,j,k) = m_ib
              endif
             enddo
          enddo
        enddo


c-----------------------------------------------------------------
c    Set "extra smoothing region" around excision region(s)
c    by surrounding inner boundary with layer after layer of ESR
c    points (number of layers given by esr_size)
c-----------------------------------------------------------------
      do loop = 1, esr_size

        do k = 2, nz-1
          do j = 2, ny-1
            do i = 2, nx-1
              if ( (mask(i,j,k) .eq. m_int)  .and. (
     &              mask(i+1,j,k) .eq. m_ib .or.
     &              mask(i-1,j,k) .eq. m_ib .or.
     &              mask(i,j+1,k) .eq. m_ib .or.
     &              mask(i,j-1,k) .eq. m_ib .or.
     &              mask(i,j,k+1) .eq. m_ib .or.
     &              mask(i,j,k-1) .eq. m_ib .or.
     &              mask(i+1,j,k) .eq. m_esr .or.
     &              mask(i-1,j,k) .eq. m_esr .or.
     &              mask(i,j+1,k) .eq. m_esr .or.
     &              mask(i,j-1,k) .eq. m_esr .or.
     &              mask(i,j,k+1) .eq. m_esr .or.
     &              mask(i,j,k-1) .eq. m_esr 
     &              ) ) then
c-------------------------------------------------------------------
c     Set to m_esr_pass (instead of m_esr) first, because otherwise
c     the thing just grows without bound in the +x, +y and +z
c     directions...
c-------------------------------------------------------------------
                  mask(i,j,k) = m_esr_pass
              endif
             enddo
          enddo
        enddo

c-------------------------------------------------------------------
c    Now switch values from m_esr_pass to m_esr
c-------------------------------------------------------------------
        do k = 2, nz-1
          do j = 2, ny-1
            do i = 2, nx-1
              if (mask(i,j,k) .eq. m_esr_pass) mask(i,j,k) = m_esr
            enddo
          enddo
        enddo
c  end of 'do loop = ', for ESR
      enddo

c-------------------------------------------------------------------
c    Set outer boundary points
c-------------------------------------------------------------------
         do k = 1, nz
            do j = 1 , ny
               mask(1, j,k) = m_ob
               mask(nx,j,k) = m_ob
            end do
         end do

         do k = 1, nz
            do i = 1 , nx
               mask(i,1,k)  = m_ob
               mask(i,ny,k) = m_ob
            end do
         end do

         do j = 1 , ny
            do i = 1 , nx
               mask(i,j,1)  = m_ob
               mask(i,j,nz) = m_ob
            end do
         end do


      return
      end

c===========================================================
c   fillm:  Fills function u (which is \tidle{u}^h) on mask
c
c   It fills the boundary of the mask with the continuum 
c   solution and fills the interior of the mask with a scalar
c   (scalar set by "exval" in maskcodes.inc)
c 
c   INPUTS:
c   h
c   nx,ny,nz
c
c   OUTPUTS:
c   u
c   mask
c===========================================================
      subroutine fillm(u,mask,x,y,z,nx,ny,nz)

         implicit       none

         include       'mgprob.inc'
         include       'maskcodes.inc'

         integer        nx,       ny,        nz
         real*8         u(nx,ny,nz)
         integer        mask(nx,ny,nz)
         real*8         x(nx),    y(ny),     z(nz)

c Local variables
         integer        i,        j,         k
         real*8         uexact


c        Inner boundary points
         call clcu2(u,x,y,z,nx,ny,nz,
     &             mask,.false.,.true.,.false.,.false.,.false.)

c        Excised points
         do  k = 1 , nz
           do  j = 1 , ny
             do  i = 1 , nx
                 if (mask(i,j,k) .eq. m_ex) u(i,j,k) = exval
             end do
           end do
         end do

         return
      end


c==================================================================
c  f2cfibfill: Fine-to-Course Inner Boundary fill
c
c  For use when one doesn't want to use the continuum solution
c  to provide data on coarse grid inner boundary points
c  Incidentally, if one does nothing to the CG IB points (modulo the
c  possible writing of "garbage" in robin3d.f), then these points
c  WILL have the exact data already on them, from previous
c  iterations of the Full Multigrid Algorithm (i.e. when the coarse
c  grids was once the "finest" grid...)
c
c  If the boundary points on the coarse & fine grid match up
c  (are in the same location), then we do a straight copy.
c  If the coarse grid boundary is interior to the fine grid boundary,
c  we use a quadratic fit of values the fine grid to load inner 
c  boundary points on the coarse grid
c
c  This is spaghetti code.  Enjoy. :-)
c
c  INPUTS:
c  uf
c  nxc, nyc, nzc
c  mask_f
c  mask_c
c
c  OUTPUTS:
c  uc
c==================================================================
      subroutine f2cibfill(uf,uc,nxc,nyc,nzc,mask_f,mask_c)
     
         implicit       none
     
         include        'mgprob.inc'
         include        'maskcodes.inc'
     
         integer        nxc,                nyc,        nzc
         real*8         uf(2 * nxc - 1, 2 * nyc - 1, 2 * nzc - 1),
     &                  uc(nxc,nyc,nzc)
         integer        mask_f(2 * nxc - 1, 2 * nyc - 1, 2*nzc-1)
         integer        mask_c(nxc,nyc,nzc)

c   Local variables
         integer        ic,       jc,       kc
         integer        if,       jf,       kf

         real*8         tmp
         integer        avgcount
         integer        di,     dj,     dk


         do kc = 2, nzc-1
           do jc = 2, nyc-1
             do ic = 2, nxc-1

                if (mask_c(ic,jc,kc) .eq. m_ib) then
                  if = 2 * ic - 1 
                  jf = 2 * jc - 1 
                  kf = 2 * kc - 1 

                  if (mask_f(if,jf,kf) .eq. m_ib) then

c--------------------------------------------------------------------
c            If CG IB point is overtop of FG IB point, just Copy
c--------------------------------------------------------------------
                     uc(ic,jc,kc) = uf(if,jf,kf)

                  else if (ic. ge. 3 .and. ic .le. nxc-2  .and.
     &                     jc. ge. 3 .and. jc .le. nyc-2  .and.
     &                     kc. ge. 3 .and. kc .le. nzc-2) then

c--------------------------------------------------------------------
c            Otherwise...
c--------------------------------------------------------------------

c   Skip this and go to another condition
c              goto 500
c--------------------------------------------------------------------
c            Dirichlet:  Quadratic Extrapolation
c            Average results of extrapolation in as many directions
c                as possible
c--------------------------------------------------------------------
                     tmp = 0.0d0
                     avgcount = 0

                     do di = -1, 1
                       do dj = -1, 1
                         do dk = -1, 1
                           if (mask_f(if+di,jf+dj,kf+dk) .ne. m_ex 
     &                       .and. mask_f(if+2*di,jf+2*dj,kf+2*dk) 
     &                       .ne. m_ex) then

                             avgcount = avgcount+1 
 
                             tmp = tmp + 
     &                       (4*uf(if+di,jf+dj,kf+dk) -
     &                          uf(if+2*di,jf+2*dj,kf+2*dk) )/3.d0

                           endif
                         end do
                       end do
                     end do

                     if (avgcount .ne. 0) then
                        uc(ic,jc,kc) = tmp/avgcount
                     else
                        write(0,*) 'f2cibfill: avgcount eq 0'
                        write(0,*) 'if,jf,kf = ',if,jf,kf
                        write(0,*) 'nxc,nyc,nzc = ',nxc,nyc,nzc
                        stop
                     endif

c SHH: check
c                     uc(ic,jc,kc) = 1.0d90 * uc(ic,jc,kc)

c                     write(0,*) '  f2cibfill: Dirichlet. ic,jc,kc = ',
c     &                   ic,jc,kc, uc(ic,jc,kc)

c                 Skip over the next section
                     goto 700

500                  continue
c--------------------------------------------------------------------
c  Greg Cook's idea:  Impose Neumann IBC at fine grid IB...
c  (My implementation of this idea does not seem to work.)
c--------------------------------------------------------------------

c                  'Poison' the data, just to make sure that, at some
c                  point below, a value gets assigned.
                     uc(ic,jc,kc) = -9.99d101

                     if (mask_f(if+1,jf,kf) .ne. m_ex .and.
     &                   mask_f(if+2,jf,kf) .ne. m_ex .and.
     &                   mask_f(if+3,jf,kf) .ne. m_ex) then
                        uc(ic,jc,kc) = uc(ic+1,jc,kc) + 3*uf(if+1,jf,kf) 
     &                               - 4*uf(if+2,jf,kf) + uf(if+3,jf,kf)
                     endif
                     if (mask_f(if-1,jf,kf) .ne. m_ex .and.
     &                   mask_f(if-2,jf,kf) .ne. m_ex .and.
     &                   mask_f(if-3,jf,kf) .ne. m_ex) then
                        uc(ic,jc,kc) = -( uc(ic-1,jc,kc) + 
     &                          3*uf(if-1,jf,kf) 
     &                        - 4*uf(if-2,jf,kf) + uf(if-3,jf,kf))
                     endif

                     if (mask_f(if,jf+1,kf) .ne. m_ex .and.
     &                   mask_f(if,jf+2,kf) .ne. m_ex .and.
     &                   mask_f(if,jf+3,kf) .ne. m_ex) then
                        uc(ic,jc,kc) = uc(ic,jc+1,kc) + 3*uf(if,jf+1,kf) 
     &                               - 4*uf(if,jf+2,kf) + uf(if,jf+3,kf)
                     endif
                     if (mask_f(if,jf-1,kf) .ne. m_ex .and.
     &                   mask_f(if,jf-2,kf) .ne. m_ex .and.
     &                   mask_f(if,jf-3,kf) .ne. m_ex) then
                        uc(ic,jc,kc) = -( uc(ic,jc-1,kc) + 
     &                                3*uf(if,jf-1,kf) 
     &                              - 4*uf(if,jf-2,kf) + uf(if,jf-3,kf))
                     endif

                     if (mask_f(if,jf,kf+1) .ne. m_ex .and.
     &                   mask_f(if,jf,kf+2) .ne. m_ex .and.
     &                   mask_f(if,jf,kf+3) .ne. m_ex) then
                        uc(ic,jc,kc) = uc(ic,jc,kc+1) + 3*uf(if,jf,kf+1) 
     &                             - 4*uf(if,jf,kf+2) + uf(if,jf,kf+3)
                     endif
                     if (mask_f(if,jf,kf-1) .ne. m_ex .and.
     &                   mask_f(if,jf,kf-2) .ne. m_ex .and.
     &                   mask_f(if,jf,kf-3) .ne. m_ex) then
                        uc(ic,jc,kc) = -( uc(ic,jc,kc-1) + 
     &                          3*uf(if,jf,kf-1) 
     &                        - 4*uf(if,jf,kf-2) + uf(if,jf,kf-3))
                     endif
                     write(0,*) 'Neumann gives ',uc(ic,jc,kc)
c                     write(0,*) '--------------'
c                     if (abs(uc(ic,jc,kc)).gt.1000) then
c              write(0,*) 'ic,jc,kc = ',ic,jc,kc
c              write(0,*) 'if,jf,kf = ',if,jf,kf
c              write(0,*) 'mask_c(ic,jc,kc) = ',mask_c(ic,jc,kc)
c              write(0,*) 'mask_f(if,jf,kf) = ',mask_f(if,jf,kf)
c              write(0,*) 'mask_f(if+1,jf,kf) = ',mask_f(if+1,jf,kf)
c              write(0,*) 'mask_f(if+2,jf,kf) = ',mask_f(if+2,jf,kf)
c              write(0,*) 'mask_f(if+3,jf,kf) = ',mask_f(if+3,jf,kf)
c              write(0,*) 'mask_f(if-1,jf,kf) = ',mask_f(if-1,jf,kf)
c              write(0,*) 'mask_f(if-2,jf,kf) = ',mask_f(if-2,jf,kf)
c              write(0,*) 'mask_f(if-3,jf,kf) = ',mask_f(if-3,jf,kf)
c                        call printia(mask_c,nxc,nyc,nzc)
c                        call printia(mask_f,2*nxc-1,2*nyc-1,2*nzc-1)
c                        stop
c                     endif

                   goto 700

600       continue

c         Zero-derivative condition
c         Not implemented yet.
             if (mask_f(if,jf,kf+1) .lt. m_ex) 
     &            uc(ic,jc,kc) = uf(if,jf,kf+1)
             if (mask_f(if,jf,kf-1) .lt. m_ex) 
     &            uc(ic,jc,kc) = uf(if,jf,kf-1)

             if (mask_f(if,jf+1,kf) .lt. m_ex) 
     &            uc(ic,jc,kc) = uf(if,jf+1,kf)
             if (mask_f(if,jf-1,kf) .lt. m_ex) 
     &            uc(ic,jc,kc) = uf(if,jf-1,kf)

             if (mask_f(if+1,jf,kf) .lt. m_ex) 
     &            uc(ic,jc,kc) = uf(if+1,jf,kf)
             if (mask_f(if-1,jf,kf) .lt. m_ex) 
     &            uc(ic,jc,kc) = uf(if-1,jf,kf)

700       continue

                  else
                    write(0,*) ' *** f2cfibfill: Error, cant set point'
                    write(0,*) ' ic, jc, kc = ', ic,jc,kc
                    write(0,*) ' if, jf, kf = ', if,jf,kf
                    write(0,*) ' mask_c, mask_f = ', 
     &                       mask_c(ic,jc,kc), mask_f(if,jf,kf)
                  endif

c           Endif for 'is this a boundary point?'
                endif
             end do
           end do
         end do


         return
         end

c===========================================================
c   exoffm:  Loads exact function everywhere BUT the mask
c
c   This is ONLY used if one wishes to `cheat'
c   (e.g. in robin3d.f)
c===========================================================
      subroutine exoffm(u,mask,x,y,z,nx,ny,nz)

         implicit       none

         include       'mgprob.inc'
         include       'maskcodes.inc'

         integer        nx,       ny,        nz
         real*8         u(nx,ny,nz)
         integer        mask(nx,ny,nz)
         real*8         x(nx),    y(ny),     z(nz)
        
         integer        i,        j,         k
         
         call clcu2(u,x,y,z,nx,ny,nz,
     &              mask,.false.,.true.,.true.,.true.,.true.)

         return

      end


c======================================================================
c     mprint: handy routine for showing the mask
c======================================================================
      subroutine mprint(mask, nx, ny, nz, ibgn,iend,jbgn,jend,kbgn,kend)
      implicit none

      integer nx, ny, nz
      integer mask(nx, ny, nz)
      integer ibgn,iend,jbgn,jend,kbgn,kend

      integer i,j,k


      write(0,*) 'Mask:'
      do k = kbgn, kend
        do j = jbgn, jend
          write(0,1000) ( mask(i,j,k), i = ibgn, iend )
        enddo
        write(0,*)
      enddo
1000  format(200I2)
      write(0,*)

      return
      end



c===========================================================
c     Load 2-D integer array with scalar.
c===========================================================
      subroutine imls(a,s,nx,ny,d3)

         implicit       none

         integer        nx,       ny,      d3
         integer        a(nx,ny,d3)
         integer        s

         integer        i,        j,       k

         do k = 1 , d3
           do j = 1 , ny
             do i = 1 , nx
               a(i,j,k) = s
             end do     
           end do
         end do

         return

      end


c===========================================================
c     Load 2-D array with scalar on mask & boundary.
c       and on first point next to boundary
c    this is handy for final output of local truncation error
c===========================================================
      subroutine lsnearm(a,s,mask,nx,ny,d3)

         implicit       none
         include       'maskcodes.inc'

         integer       nx,       ny,      d3
         real*8        a(nx,ny,d3)
         real*8        s
         integer       mask(nx,ny,d3)

         integer        i,        j,       k
        

         do k = 1 , d3
           do j = 1 , ny
             do i = 1 , nx
               if    ((mask(i,j,k)   .ge. m_ex)
     &            .or.(mask(i,j,k+1) .ge. m_ex )
     &            .or.(mask(i,j,k-1) .ge. m_ex )
     &            .or.(mask(i,j+1,k) .ge. m_ex )
     &            .or.(mask(i,j-1,k) .ge. m_ex )
     &            .or.(mask(i+1,j,k) .ge. m_ex )
     &            .or.(mask(i-1,j,k) .ge. m_ex )
     &
     &            .or.(mask(i,j+1,k+1) .ge. m_ex )
     &            .or.(mask(i,j+1,k-1) .ge. m_ex )
     &            .or.(mask(i,j-1,k+1) .ge. m_ex )
     &            .or.(mask(i,j-1,k-1) .ge. m_ex )
     &            .or.(mask(i+1,j,k+1) .ge. m_ex )
     &            .or.(mask(i+1,j,k-1) .ge. m_ex )
     &            .or.(mask(i-1,j,k+1) .ge. m_ex )
     &            .or.(mask(i-1,j,k-1) .ge. m_ex )
     &            .or.(mask(i+1,j+1,k) .ge. m_ex )
     &            .or.(mask(i+1,j-1,k) .ge. m_ex )
     &            .or.(mask(i-1,j+1,k) .ge. m_ex )
     &            .or.(mask(i-1,j-1,k) .ge. m_ex )
     &            )  then
                   a(i,j,k) = s
                  else 
               endif

             end do
           end do
         end do

         return

      end

c===========================================================
c      Check for a scalar value in array.  Used to check
c      for 'poisoned' data
c===========================================================
      subroutine checkfor(a,s,nx,ny,nz,mask,wholegrid)
         
      implicit       none

      include        'maskcodes.inc'
           
      integer        nx,       ny,      nz
      real*8         a(nx,ny,nz)
      real*8         s
      integer        mask(nx,ny,nz) 
      logical        wholegrid  

      integer        i,        j,       k
               
      do k = 1 , nz
         do j = 1 , ny
            do i = 1 , nx
               if (abs(a(i,j,k)) .ge. abs(s) .and.
     &             (wholegrid .or. mask(i,j,k) .ne. m_ex) ) then
                    write(0,*) 'checkfor: true at ',i,j,k
                    write(0,*) ' nx, ny, nz = ',nx,ny,nz
                    write(0,*) ' a(i,j,k) = ',a(i,j,k)
                    write(0,*) ' mask(i,j,k) = ',mask(i,j,k)
                    stop
               endif
            end do
         end do
      end do
      
      return
     
      end


