c================================================================
c   robin3d                                          Scott Hawley
c                                     shawley@physics.utexas.edu
c
c    'robin3d' and supporting routines constitute a complete 
c    FAS multi-grid solver for
c
c    u_xx + u_yy + u_zz + G(u) = rhs(x,y,z)       (1)
c
c    on a unit cube of domain [-0.5,-0.5,-0.5] to [0.5,0.5,0.5],
c    with one or more holes (excised regions) in the interior of 
c    the domain.   The inner boundary is specified via a Dirchlet
c    condition using the continuum solution (see below), and the 
c    outer boundary is subject a Robin condition.
c
c     Currently program solves (1) with
c
c        G(u) = sigma * u*u
c              or
c        G(u) = - K^2 u^5 + A^2 / u^7
c
c     and 'rhs' specified so that
c
c        u = 1 - 2 M / r
c
c     where M, sigma, K^2 and A^2 are set 'mgprob.inc'.
c
c  This code has been modified from Matthew Choptuik's mgfas2d code, 
c  which solves a nonlinear Poisson equation in 2D without excision
c
c  Note: robin3d is a subroutine called from cwrapper.c, which
c  allocates memory initially.  (C offers a bit more control of
c  memory allocation than does Fortran (i.e. whether to place 
c  storage on the heap vs. the stack) 
c
c  INPUTS:
c  q          Big array from which smaller arrays are "allocated"
c  qsize      Number of elements in q
c  level      Maximum multigrid level to use (2^level+1 points per edge)
c  lmin       Lowest multigrid level to use (2^lmin+1 points per edge)
c  mrad       Physical Radius of mask on finest grid.
c                Negative value turns excision off.
c  mdisp_x    Displacement in x-direction of each exicision region
c  in_sigma   sigma parameter in NL Poisson eq
c  nvcycle    # of V-cycles to perform
c  preswp     # of Pre-CGC smoothing sweeps
c  pstswp     # of Post-CGC smoothing sweeps
c  in_gft_trace  Flag to turn on/off .sdf output files
c
c  OUTPUTS:
c  Output to stderr, stdout and to files (.sdf and fort.*)
c================================================================
      subroutine   robin3d(
     &                q, qsize, level, lmin,
     &                mrad, mdisp_x, 
     &                in_sigma, 
     &                nvcycle, preswp, pstswp,
     &                in_gft_trace,
     &                in_eqtype, in_excttype,
     &                in_M, in_Ksq, in_Asq, 
     &                in_maxxyz, in_minxyz
     &             )

      implicit     none

c-----------------------------------------------------------
c     Functions called
c-----------------------------------------------------------
      real*8      dmnrm2

c-----------------------------------------------------------
c     Passed-in parameters
c-----------------------------------------------------------
      integer      qsize
      real*8       q(qsize)
      integer      level,      nvcycle,    preswp,    pstswp
      integer      lmin
      real*8       mrad        
      real*8       mdisp_x
      real*8      in_M
      real*8      in_sigma,    in_Ksq,    in_Asq
      integer     in_gft_trace
      integer     in_eqtype,   in_excttype
      real*8      in_maxxyz,    in_minxyz

c-----------------------------------------------------------
c     'mg.inc' contains the main 'data structure' used 
c     in this implementation.
c-----------------------------------------------------------
      include     'mg.inc'


c-----------------------------------------------------------
c     'mgprob.inc' contains problem-specific parameters
c     including command-line arguments 'sigma' and 
c     'gft_trace'.
c-----------------------------------------------------------
      include     'mgprob.inc'

c-----------------------------------------------------------
c     Locals.
c-----------------------------------------------------------
      integer      l,          cycle
      real*8       norm
      logical      ltrace
      parameter  ( ltrace = .false. )



c-----------------------------------------------------------
c     Parse input params
c-----------------------------------------------------------
      eqtype = in_eqtype
      excttype = in_excttype
      M = in_M
      sigma = in_sigma
      Ksq = in_Ksq
      Asq = in_Asq
      xyzmax = in_maxxyz
      xyzmin = in_minxyz
      gft_trace =  (in_gft_trace .eq. 1)
      if( gft_trace ) write(0,*) 'gft tracing enabled'

c SHH: Include following two lines for "scale invariance" check:
c     such that one can run with same command line, for various M,
c     and always get the same results.
c      mrad = mrad * M
c      sigma = sigma / M / M

c-----------------------------------------------------------
c     Initialize data structure.  All grid function arrays
c     are allocated from global array 'q' (see 'mg.inc')
c-----------------------------------------------------------
      call gsinit(q,qsize,level,lmin)

c-----------------------------------------------------------
c     Set up the mask
c-----------------------------------------------------------
      call minit(level,lmin,mrad,mdisp_x,q,qsize)
 
c-----------------------------------------------------------
c     Solve series of problems from l = lmin through 
c     l = level.
c-----------------------------------------------------------
      do l = lmin , level
         if( l .eq. lmin ) then
c-----------------------------------------------------------
c           Initialize level lmin solution (coarsest) 
c           with zeros
c-----------------------------------------------------------
            call dmls  ( q(u(l)), 1.0d0, nx(l), ny(l), nz(l) )

            call fillm( q(u(l)), qm(mask(l)), 
     &          q(x(l)), q(y(l)), q(z(l)), nx(l), ny(l), nz(l))

cSHH    If you want to cheat: initialize with exact solution
c            call clcu2( q(u(l)), q(x(l)), q(y(l)), q(z(l)), 
c     &                   nx(l), ny(l), nz(l), qm(mask(l)), 
c     &                   .false., .true., .true., .true., .true.)

c            call printslice( q(u(l)), 
c     &           q(x(l)), q(y(l)), q(z(l)),
c     &           nx(l), ny(l), nz(l),
c     &          l,level, .false. )

         else
c-----------------------------------------------------------
c           Initialize level l soln via linear interpolation
c           of level l-1 soln.
c-----------------------------------------------------------
            call fillm( q(u(l-1)), qm(mask(l-1)), 
     &                  q(x(l-1)), q(y(l-1)), q(z(l-1)),
     &                  nx(l-1), ny(l-1), nz(l-1))

            call dmlint ( q(u(l-1)), q(u(l)), 
     &                    nx(l-1), ny(l-1), nz(l-1) )

c SHH: check that no excised values get interpolated into normal
c      part of grid... Depening on how IB is defined, i.e. if
c      "diagonal" points are set as IB points in maskutils.c, this
c      check may be important
            call checkfor(q(u(l)), 1.0d10, nx(l), ny(l), nz(l),
     &                         qm(mask(l)), .false. )

            call robinbc(q(u(l)), nx(l), ny(l), nz(l), 
     &                   q(x(l)), q(y(l)), q(z(l)), qm(mask(l)))

            call fillm( q(u(l)), qm(mask(l)), 
     &                  q(x(l)), q(y(l)), q(z(l)), 
     &                  nx(l), ny(l), nz(l))

c    If you want to cheat: initialize with exact solution
c            call exoffm( q(u(l)), qm(mask(l)), 
c     &          q(x(l)), q(y(l)), q(z(l)), nx(l), ny(l), nz(l))

         end if

c-----------------------------------------------------------
c        Initialize 'rhs' (rhs = f^h).
c-----------------------------------------------------------
         call clcrhs2   ( q(rhs(l)), q(x(l)), q(y(l)), q(z(l)), 
     &                    nx(l), ny(l), nz(l) )

c-----------------------------------------------------------
c        "Solve" level l problem by applying cycle 
c        (preswp,pstswp) V-cycles.
c-----------------------------------------------------------
         do cycle = 1 , nvcycle
            call vcycle(q,qsize,l,lmin,cycle,preswp,pstswp)
         end do

c-----------------------------------------------------------
c   output a slice through the solution itself
c-----------------------------------------------------------
c            call printslice( q(u(l)), 
c     &           q(x(l)), q(y(l)), q(z(l)),
c     &           nx(l), ny(l), nz(l),
c     &          l,level, .false. )


c-----------------------------------------------------------
c        Compute relative truncation error
c        (tau should already have been calculated in vcycle, but
c        we don't take any chances...)
c-----------------------------------------------------------
         if (l .gt. lmin) then
            call calctau(q,qsize,l)
            call lsnearm(q(tau(l-1)),0.0d0,qm(mask(l-1)),
     &           nx(l-1),ny(l-1),nz(l-1))
            norm = dmnrm2( q(tau(l-1)), nx(l-1), ny(l-1), nz(l-1) )

c        Make the norm of tau a dimensionless quantity
            norm = norm * M * M
            write(0,*) 'l(h), || tau_h^2h|| = ', l,norm
         endif

c-----------------------------------------------------------
c        Compute norm of error on finest gird and output to 
c        standard error.
c-----------------------------------------------------------
         call calcse(q(u(l)), q(t1(l)), qm(mask(l)), 
     &          q(x(l)), q(y(l)), q(z(l)), nx(l), ny(l), nz(l))

         norm = dmnrm2( q(t1(l)), nx(l), ny(l), nz(l) )

            call printslice( q(t1(l)), 
     &           q(x(l)), q(y(l)), q(z(l)),
     &           nx(l), ny(l), nz(l),
     &          l,level, .false. )


         write(0,1000) l,norm
         write(*,1000) l,norm
1000     format (i4,e25.16)

      end do
      l = level
      call gftout(q(t1(l)),'err',nx(l),ny(l), nz(l))

c-----------------------------------------------------------
c     Close RNPL-style .hdf files if option enabled.
c-----------------------------------------------------------
      if( gft_trace ) then
c       Don't have RNPL libraries properly installed on Mac.
c         call gft_close_all()
      end if

c-----------------------------------------------------------
c     Normal exit.
c-----------------------------------------------------------
      return

      end




c===========================================================
c  vcycle:
c     Performs one FAS V-cycle of level 'l' problem, using 
c     'preswp' pre-CGC and 'pstswp' pst-CGC sweeps.  
c     Supply cycle=N for Nth consecutive invocation of this 
c     routine on the same level.
c===========================================================
      subroutine vcycle(q,qsize,l,lmin,cycle,preswp,pstswp)
 
         implicit       none 
 
         include       'mg.inc'

         integer        qsize
         real*8         q(qsize)
         integer        l,        lmin
         integer        cycle,    preswp,   pstswp
 
c  Local variables
         real*8         dmnrm2,   relaxc3

         real*8         norm
         integer        m,        sweep
         real*8         rresnrm

c    Switch to toggle whether we apply the outer BC on all grid
c      levels or only on the finest level -- should be set to true!
         logical        bconall
         parameter    ( bconall = .true. )
 

c-----------------------------------------------------------
c        Work from fine level to coarse level, performing
c        pre-CGC sweeps and intitiating coarse grid 
c        corrections.  Skipped if on coarsest level, l = lmin.
c-----------------------------------------------------------
         do m = l , lmin+1 , -1

c-----------------------------------------------------------
c           Pre-CGC relaxation sweeps.
c-----------------------------------------------------------

            if ( cycle .eq. 1  .or.  m .ne. l ) then

               if (m.eq.l .or. bconall) 
     &             call robinbc(q(u(m)), nx(m), ny(m), nz(m),
     &                          q(x(m)), q(y(m)), q(z(m)), qm(mask(m)))

               do sweep = 1 , preswp

c   Relax an extra time in the "extra smoothing region"
                  rresnrm = relaxc3(q(u(m)),q(rhs(m)),
     &                              nx(m),ny(m),nz(m),h(m),
     &                              qm(mask(m)), .false.)
                  call gftout(q(u(m)),'urc',nx(m),ny(m),nz(m))


c   The "real" relaxation
                  rresnrm = relaxc3(q(u(m)),q(rhs(m)),
     &                              nx(m),ny(m),nz(m),h(m),
     &                              qm(mask(m)), .true. )

                  if (m.eq.l .or. bconall) 
     &               call robinbc(q(u(m)),nx(m),ny(m),nz(m),
     &                           q(x(m)), q(y(m)), q(z(m)), qm(mask(m)))

                  call gftout(q(u(m)),'urc',nx(m),ny(m),nz(m))

                  call gftout(q(u(m)),'u',nx(m),ny(m),nz(m))
                  call trace(m,'Pre-CGC',rresnrm)
               end do
            end if

c-----------------------------------------------------------
c           Compute truncation error estimate.
c
c               2h      2h 2h h     2h h  h
c           tau    :=  L  I  u  -  I  L  u
c               h         h        h
c-----------------------------------------------------------
            call calctau(q,qsize,m)

c  For testing only:
c        To make sure we're not using exact data on the 2h grid,
c            write in garbage for rhs^2h and u^2h
            call dmls ( q(rhs(m-1)), -1.0d83, nx(m-1), ny(m-1), nz(m-1))
            call dmls ( q(u(m-1)), -1.0d83, nx(m-1), ny(m-1), nz(m-1))

c-----------------------------------------------------------
c
c               2h       2h    h       2h
c           rhs     :=  I   rhs   + tau
c                        h              h
c
c            2h         2h  h
c           u       := I   u
c                       h
c-----------------------------------------------------------
            call dmrshw ( q(rhs(m)), q(rhs(m-1)), nx(m-1), ny(m-1),
     &                    nz(m-1), qm(mask(m)), qm(mask(m-1)), .false.)
c            call checkfor(q(rhs(m-1)),-1.0d83,nx(m-1),ny(m-1), nz(m-1),
c     &                    qm(mask(m-1)),.false.)

            call dmma   ( q(rhs(m-1)), q(tau(m-1)), q(rhs(m-1)),
     &                    nx(m-1), ny(m-1), nz(m-1) )


            call dmrshw ( q(u(m)), q(u(m-1)), nx(m-1), ny(m-1), nz(m-1),
     &                    qm(mask(m)), qm(mask(m-1)), .false.)
c            call checkfor(q(u(m-1)),-1.0d83,nx(m-1),ny(m-1), nz(m-1),
c     &                    qm(mask(m-1)), .false.)


c-----------------------------------------------------------
c           Continue loop to perform coarse-grid 
c           correction.
c-----------------------------------------------------------
         end do

c-----------------------------------------------------------
c        Solve coarse-grid problem using relaxation.
c        Note use of 'do while()'
c-----------------------------------------------------------
         rresnrm = 1.0d0 + epsi1
         sweep = 1

         if (lmin.eq.l .or. bconall) 
     &      call robinbc(q(u(lmin)),nx(lmin),ny(lmin),nz(lmin),
     &                   q(x(lmin)), q(y(lmin)), q(z(lmin)), 
     &                   qm(mask(lmin)))

         do while( 
     &             rresnrm .gt. epsi1  .and. 
     &             sweep .le. maxsweep1 )

c            call printslice( q(u(lmin)), 
c     &           q(x(lmin)), q(y(lmin)), q(z(lmin)),
c     &           nx(lmin), ny(lmin), nz(lmin),
c     &          lmin,l, .false. )

            rresnrm = relaxc3( q(u(lmin)), q(rhs(lmin)),
     &                         nx(lmin), ny(lmin), nz(lmin), h(lmin),
     &                         qm(mask(lmin)), .false.)

            rresnrm = relaxc3( q(u(lmin)), q(rhs(lmin)),
     &                         nx(lmin), ny(lmin), nz(lmin), h(lmin),
     &                         qm(mask(lmin)), .true.)
            if (lmin.eq.l .or. bconall) 
     &         call robinbc(q(u(lmin)),nx(lmin),ny(lmin),nz(lmin),
     &                   q(x(lmin)), q(y(lmin)), q(z(lmin)), 
     &                   qm(mask(lmin)))
            
            call gftout(q(u(lmin)),'u',nx(lmin),ny(lmin),nz(lmin))
            call trace(lmin,'Coarse grid',rresnrm)
            sweep = sweep + 1
         end do

         if( rresnrm .gt. epsi1 ) then
            write(0,*) '*** ERROR vcycle: Coarse grid '//
     &                 'relaxation failed.'
            stop
         end if


c-----------------------------------------------------------
c        Now work from coarse to fine level, updating 
c        unknowns and performing post-CGC relaxtion 
c        sweeps.  Skipped if l = lmin.
c-----------------------------------------------------------
         do m = lmin+1 , l
c-----------------------------------------------------------
c           Linearly interpolate correction and update 
c           level m unknown.
c
c            h       h    h    2h     2h  h
c           u    := u  + I  ( u   -  I   u  )
c                         2h          h
c-----------------------------------------------------------

            call dmrshw ( q(u(m)), q(t1(m-1)), nx(m-1), ny(m-1),
     &                    nz(m-1),qm(mask(m)), qm(mask(m-1)), .false.)

            call dmms   ( q(u(m-1)), q(t1(m-1)), q(t1(m-1)),
     &                    nx(m-1), ny(m-1), nz(m-1) )

            call dmlint ( q(t1(m-1)), q(t1(m)),
     &                    nx(m-1),ny(m-1), nz(m-1) )


            call dmma   ( q(u(m)), q(t1(m)), q(u(m)),
     &                    nx(m),ny(m), nz(m) )

c-----------------------------------------------------------
c           Post-CGC relaxation sweeps
c-----------------------------------------------------------
            if (m .eq. l) then
              call fillm  ( q(u(m)), qm(mask(m)), 
     &           q(x(m)), q(y(m)), q(z(m)), nx(m), ny(m), nz(m))
            endif

            if (m.eq.l .or. bconall) 
     &         call robinbc(q(u(m)),nx(m),ny(m),nz(m), 
     &                      q(x(m)), q(y(m)), q(z(m)), qm(mask(m)))

            do sweep = 1 , pstswp

               rresnrm = relaxc3 ( q(u(m)), q(rhs(m)),
     &                             nx(m),ny(m),nz(m),h(m), 
     &                             qm(mask(m)), .false.)
               rresnrm = relaxc3 ( q(u(m)), q(rhs(m)),
     &                             nx(m),ny(m),nz(m),h(m), 
     &                             qm(mask(m)), .true.)

               if (m.eq.l .or. bconall)
     &            call robinbc(q(u(m)),nx(m),ny(m),nz(m),
     &                      q(x(m)), q(y(m)), q(z(m)), qm(mask(m)))

               call gftout(q(u(m)),'u',nx(m),ny(m),nz(m))
               call trace(m,'Post-CGC',rresnrm)
            end do

          end do

c-----------------------------------------------------------
c         End of the V-cycle.
c-----------------------------------------------------------
         write(0,*)

         return
      end

