
c         SUPPORTING ROUTINES FOR robin3d
c         Code is similar to Matt Choptuik's mgfas2dlib.f,
c         with additional subroutines added

c===========================================================
c     Performs red-black GS relaxation sweep of 2d model 
c     problem and returns norm of running ('instantaneous') 
c     residuals.
c     this is for the equation
c      del^2 u + sigma u^2 = rhs
c===========================================================
      double precision function relaxc3(u,rhs,nx,ny,nz,h,
     &                 mask, norm_int_too)

      implicit       none

      include       'mgprob.inc'
      include       'maskcodes.inc'
 
      double precision rel_usq, rel_constr

      integer        nx,              ny,               nz
      real*8         u(nx,ny,nz),     rhs(nx,ny,nz)
      real*8         h
      integer        mask(nx,ny,nz)

c     Flag. ESR is always smoothed, normal interior only smoothed if
c       norm_int_too is true
      logical        norm_int_too

        if (eqtype .eq. 1) then
          relaxc3 =  rel_usq(u,rhs,nx,ny,nz,h, mask, norm_int_too)
        else if (eqtype .eq. 2) then
          relaxc3 =  rel_constr(u,rhs,nx,ny,nz,h, mask, norm_int_too)
        else
           write(0,*) 'relaxc3: Error, unknown eqtype = ',eqtype
           stop
        endif

      return

      end

c===========================================================
c     Performs red-black GS relaxation sweep of 2d model 
c     problem and returns norm of running ('instantaneous') 
c     residuals.
c     this is for the equation
c      del^2 u + sigma u^2 = rhs
c===========================================================
      double precision function rel_usq(u,rhs,nx,ny,nz,h,
     &                 mask, norm_int_too)

      implicit       none

      include       'mgprob.inc'
      include       'maskcodes.inc'

      integer        nx,              ny,               nz
      real*8         u(nx,ny,nz),     rhs(nx,ny,nz)
      real*8         h
      integer        mask(nx,ny,nz)

c     Flag. ESR is always smoothed, normal interior only smoothed if
c       norm_int_too is true
      logical        norm_int_too

c   Local Variables
      real*8         hm2,      m6hm2,    rresl
      integer        i,        j,        k,
     &               isw,      jsw,      ksw,            pass
      real*8         resdenom

      integer        numpoints

      numpoints = 0

      rel_usq =  0.0d0
      hm2     =  1.0d0 / (h * h)
      m6hm2   = -6.0d0 * hm2
c      write(0,*) 'm6hm2 = ',m6hm2

c       Note "pass" and "*sw" variables are used to impose the
c       red-black ordering...
      ksw = 1
      do pass = 1 , 2
         jsw = ksw
         do k = 2, nz - 1
             isw = jsw
             do j = 2 , ny - 1
                do i = isw + 1 , nx - 1, 2
c----------------------------------------------------------------------
c     If the point is not excised, relax normally
c----------------------------------------------------------------------
                   numpoints = numpoints + 1
                   if ( ((mask(i,j,k) .eq. m_int) .and. norm_int_too) 
     &                 .or. (mask(i,j,k) .eq. m_esr) ) then
                       rresl = hm2 * (u(i+1,j,k) + u(i-1,j,k) 
     &                              + u(i,j+1,k) + u(i,j-1,k)
     &                              + u(i,j,k+1) + u(i,j,k-1) 
     &                              - 6.0d0 * u(i,j,k)) +
     &                                sigma * u(i,j,k)**2 -
     &                                rhs(i,j,k)
c             Following line for use with "relative" residual norm
c             Omit if you want to use an "absolute" residual norm
c          (Relative norm needed for checking "scaling by M" of problem)
c                        rresl = rresl / (m6hm2 + 2.0d0*sigma *u(i,j,k))
c

                        u(i,j,k) = u(i,j,k) - rresl  
c             Following line omitted for "relative" residual norm, 
c                included for absolute norm
     &                      / (m6hm2 + 2.0d0 * sigma * u(i,j,k))


c Richard's idea: perhaps denominator of Newton-Gauss-Seidel iteration
c becomes small as radius increases (for u = 1 + 2 M /r), and 
c perhaps this is the cause of...everything screwing up.  Wrote this
c code to check this.  
                        resdenom = m6hm2 + 2.0d0 * sigma * u(i,j,k)
                        if (abs(resdenom) .lt. 1.0d0) then
                        write(0,*) 'resdenom = ',resdenom
                           write(0,*) 'pos at i, j, k = ',i,j,k
                           stop
                        endif

                        rel_usq = rel_usq + rresl * rresl
                    else
c----------------------------------------------------------------------
c      If the point is excised, do nothing!
c----------------------------------------------------------------------
                    endif
                  end do
                isw = 3 - isw
             end do
             jsw = 3 - jsw
         end do
         ksw = 3 - ksw

      enddo

      if (numpoints .gt. 0) then
            rel_usq = sqrt(rel_usq / (numpoints))
      else
            rel_usq = 0.0d0
      endif

      return

      end

c===========================================================
c     Performs red-black GS relaxation sweep of 2d model 
c     problem and returns norm of running ('instantaneous') 
c     residuals.
c     this is for the constraint-like equation
c      del^2 u - K^2 u^5 + A^2 u^(-7) = rhs
c===========================================================
      double precision function rel_constr(u,rhs,nx,ny,nz,h,
     &                 mask, norm_int_too)

      implicit       none

      include       'mgprob.inc'
      include       'maskcodes.inc'

      integer        nx,              ny,               nz
      real*8         u(nx,ny,nz),     rhs(nx,ny,nz)
      real*8         h
      integer        mask(nx,ny,nz)

c     Flag. ESR is always smoothed, normal interior only smoothed if
c       norm_int_too is true
      logical        norm_int_too

c   Local Variables
      real*8         hm2,      m6hm2,    rresl
      integer        i,        j,        k,
     &               isw,      jsw,      ksw,            pass
      real*8         delta,    denom

      integer        numpoints

      numpoints = 0

      rel_constr =  0.0d0
      hm2     =  1.0d0 / (h * h)
      m6hm2   = -6.0d0 * hm2
c      write(0,*) 'm6hm2 = ',m6hm2

c       Note "pass" and "*sw" variables are used to impose the
c       red-black ordering...
      ksw = 1
      do pass = 1 , 2
         jsw = ksw
         do k = 2, nz - 1
             isw = jsw
             do j = 2 , ny - 1
                do i = isw + 1 , nx - 1, 2
c----------------------------------------------------------------------
c     If the point is not excised, relax normally
c----------------------------------------------------------------------
                   numpoints = numpoints + 1
                   if ( ((mask(i,j,k) .eq. m_int) .and. norm_int_too) 
     &                 .or. (mask(i,j,k) .eq. m_esr) ) then
                       rresl = hm2 * (u(i+1,j,k) + u(i-1,j,k) 
     &                              + u(i,j+1,k) + u(i,j-1,k)
     &                              + u(i,j,k+1) + u(i,j,k-1) 
     &                              - 6.0d0 * u(i,j,k)) 
     &                              - Ksq * u(i,j,k)**5 
     &                              + Asq / u(i,j,k)**7 
     &                              - rhs(i,j,k)

                        denom = (m6hm2 - 5.0d0 * Ksq * u(i,j,k)**4
     &                               - 7.0d0 * Asq / u(i,j,k)**8 )

                        delta = rresl / denom 

c             Following line for use with "relative" residual norm
c             Omit if you want to use an "absolute" residual norm
c          (Relative norm needed for checking "scaling by M" of problem)
c                        rresl = rresl / denom

c                         if (i.eq.4.and.j.eq.4.and.k.eq.4) then
c                             write(0,*) '--------------'
c                             write(0,*) 'u, delta = ',
c     &                           u(i,j,k), delta
c                         endif

                        u(i,j,k) = u(i,j,k) - delta

c Richard's idea: perhaps denominator of Newton-Gauss-Seidel iteration
c becomes small as radius increases (for u = 1 + 2 M /r), and 
c perhaps this is the cause of...everything screwing up.  Wrote this
c code to check this.  
                        if (abs(denom) .lt. 1.0d0) then
                        write(0,*) 'denom = ',denom
                           write(0,*) 'pos at i, j, k = ',i,j,k
                           stop
                        endif
                     
c                         if (i.eq.4.and.j.eq.4.and.k.eq.4) then
c                             write(0,*) 'u, delta  = ',
c     &                           u(i,j,k),delta
c                         endif

                        rel_constr = rel_constr + rresl * rresl
                    else
c----------------------------------------------------------------------
c      If the point is excised, do nothing!
c----------------------------------------------------------------------
                    endif
                  end do
                isw = 3 - isw
             end do
             jsw = 3 - jsw
         end do
         ksw = 3 - ksw

      enddo

      if (numpoints .gt. 0) then
            rel_constr = sqrt(rel_constr / (numpoints))
      else
            rel_constr = 0.0d0
      endif

      return

      end


c===========================================================
c    Computes
c               h  h
c      lopu := L  u
c===========================================================
      subroutine lop(lopu,u,mask,nx,ny,nz,h)
 
         implicit       none

         include       'mgprob.inc'
         include       'maskcodes.inc'
 
         integer        nx,               ny,            nz
         real*8         lopu(nx,ny,nz),   u(nx,ny,nz)
         integer        mask(nx,ny,nz)
         real*8         h
 
         integer        i,                 j,             k
         real*8         hm2
 
         hm2 = 1.0d0 / (h * h)

         if (eqtype .eq. 1) then
           do k = 2, nz - 1
             do j = 2 , ny - 1
                do i = 2 , nx - 1
                  if (mask(i,j,k) .lt. m_ex) then
                  lopu(i,j,k) = hm2 * (u(i+1,j,k) + u(i-1,j,k) 
     &                               + u(i,j+1,k) + u(i,j-1,k)  
     &                               + u(i,j,k+1) + u(i,j,k-1)  
     &                               - 6.0d0 * u(i,j,k)) +
     &                                 sigma * u(i,j,k)**2
                 else

c               Zero out excised and inner boundary points
                    lopu(i,j,k) = 0.0d0

                 endif
               end do
             end do
           end do
         else if (eqtype .eq. 2) then
           do k = 2, nz - 1
             do j = 2 , ny - 1
                do i = 2 , nx - 1
                  if (mask(i,j,k) .lt. m_ex) then
                  lopu(i,j,k) = hm2 * (u(i+1,j,k) + u(i-1,j,k)
     &                               + u(i,j+1,k) + u(i,j-1,k)
     &                               + u(i,j,k+1) + u(i,j,k-1)
     &                               - 6.0d0 * u(i,j,k)) 
     &                               - Ksq * u(i,j,k)**5
     &                               + Asq / u(i,j,k)**7
                 else

c               Zero out excised and inner boundary points
                    lopu(i,j,k) = 0.0d0

                 endif
               end do
             end do
           end do
 
         else
            write(0,*) 'lop: Error, unknown eqtype = ',eqtype
            stop
         endif



         call dmlbs(lopu,0.0d0,nx,ny,nz)

         return
      end


c===========================================================
c     2--D half-weighted 2:1 restriction with injection
c     of border values 
c    inside mask, values are simply injected
c===========================================================
      subroutine dmrshw(uf,uc,nxc,nyc,nzc,mask_f,mask_c, dontcheck)

         implicit       none

         include        'maskcodes.inc'
         include        'mgprob.inc'

         integer        nxc,                nyc,        nzc
         real*8         uf(2 * nxc - 1, 2 * nyc - 1, 2 * nzc - 1),
     &                  uc(nxc,nyc,nzc)
         integer        mask_f(2 * nxc - 1, 2 * nyc - 1, 2*nzc-1)
         integer        mask_c(nxc,nyc,nzc)
         logical        dontcheck

         integer        ic,       jc,       if,       jf
         integer        kc,       kf
         logical        localflag 


         localflag = dontcheck
         localflag = .false.



c-------------------------------------------------------------------
c   Normal Interior points
c-------------------------------------------------------------------
         do kc = 2, nzc - 1
           kf = 2 * kc - 1
           do jc = 2 , nyc - 1
             jf = 2 * jc - 1
             do ic = 2 , nxc - 1
               if = 2 * ic - 1

c              Rules for non-boundary points
               if (mask_c(ic,jc,kc) .lt. m_ib ) then
                  if (.not. localflag .and.
     &                (
     &                     mask_f(if,jf,kf  ) .lt. m_ex
     &               .and. mask_f(if,jf,kf+1) .lt. m_ex
     &               .and. mask_f(if,jf,kf-1) .lt. m_ex
     &               .and. mask_f(if,jf+1,kf) .lt. m_ex
     &               .and. mask_f(if,jf-1,kf) .lt. m_ex
     &               .and. mask_f(if+1,jf,kf) .lt. m_ex
     &               .and. mask_f(if-1,jf,kf) .lt. m_ex
     &
     &               .and. mask_f(if,jf+1,kf+1) .lt. m_ex
     &               .and. mask_f(if,jf+1,kf-1) .lt. m_ex
     &               .and. mask_f(if,jf-1,kf+1) .lt. m_ex
     &               .and. mask_f(if,jf-1,kf-1) .lt. m_ex
     &               .and. mask_f(if+1,jf,kf+1) .lt. m_ex
     &               .and. mask_f(if+1,jf,kf-1) .lt. m_ex
     &               .and. mask_f(if-1,jf,kf+1) .lt. m_ex
     &               .and. mask_f(if-1,jf,kf-1) .lt. m_ex
     &               .and. mask_f(if+1,jf+1,kf) .lt. m_ex
     &               .and. mask_f(if+1,jf-1,kf) .lt. m_ex
     &               .and. mask_f(if-1,jf+1,kf) .lt. m_ex
     &               .and. mask_f(if-1,jf-1,kf) .lt. m_ex
     &
     &               .and. mask_f(if+1,jf+1,kf+1) .lt. m_ex
     &               .and. mask_f(if+1,jf+1,kf-1) .lt. m_ex
     &               .and. mask_f(if+1,jf-1,kf+1) .lt. m_ex
     &               .and. mask_f(if+1,jf-1,kf-1) .lt. m_ex
     &               .and. mask_f(if-1,jf+1,kf+1) .lt. m_ex
     &               .and. mask_f(if-1,jf+1,kf-1) .lt. m_ex
     &               .and. mask_f(if-1,jf-1,kf+1) .lt. m_ex
     &               .and. mask_f(if-1,jf-1,kf-1) .lt. m_ex)
     &              ) then
                    uc(ic,jc,kc) = 
     &                   0.125d0 * (
     &                      uf(if,jf,kf) 
     &                    +(uf(if+1,jf,kf) + uf(if-1,jf,kf) 
     &                    + uf(if,jf+1,kf) + uf(if,jf-1,kf) 
     &                    + uf(if,jf,kf+1) + uf(if,jf,kf-1)
     &                       ) / 2
     &                    +(uf(if,jf+1,kf+1) + uf(if,jf-1,kf+1) 
     &                    + uf(if,jf+1,kf-1) + uf(if,jf-1,kf-1)
     &                    + uf(if+1,jf,kf+1) + uf(if-1,jf,kf+1) 
     &                    + uf(if+1,jf,kf-1) + uf(if-1,jf,kf-1)
     &                    + uf(if+1,jf+1,kf) + uf(if-1,jf+1,kf)
     &                    + uf(if+1,jf-1,kf) + uf(if-1,jf-1,kf)
     &                       ) / 4
     &                    +(uf(if+1,jf+1,kf+1) + uf(if-1,jf+1,kf+1) 
     &                    + uf(if+1,jf-1,kf+1) + uf(if-1,jf-1,kf+1)
     &                    + uf(if+1,jf+1,kf-1) + uf(if-1,jf+1,kf-1)
     &                    + uf(if+1,jf-1,kf-1) + uf(if-1,jf-1,kf-1)
     &                       ) / 8
     &                     )
                   else
                       uc(ic,jc,kc) = uf(if,jf,kf)
                   endif
                else  
c                Excised and inner boundary points
                       uc(ic,jc,kc) = uf(if,jf,kf)
                endif

            end do
           end do
         end do

c---------------------------------------------------------------------
c       Along outer boundaries, just do straight injection
c---------------------------------------------------------------------


c----------------
c     x faces
c----------------
         do kc = 1, nzc
            kf = 2 * kc - 1
            do jc = 1 , nyc
               uc(1  ,jc,kc) = uf(1,  2 * jc - 1, kf)
               uc(nxc,jc,kc) = uf(2 * nxc - 1, 2 * jc - 1, kf)
            end do
         end do

c----------------
c     y faces
c----------------
         do kc = 1, nzc
            kf = 2 * kc - 1
            do ic = 1 , nxc
               uc(ic,1  ,kc) = uf(2 * ic - 1, 1, kf  )
               uc(ic,nyc,kc) = uf(2 * ic - 1, 2 * nyc - 1, kf)
            end do
         end do
   


c----------------
c      z faces
c----------------
         do jc = 1 , nyc
            jf = 2 * jc - 1
            do ic = 1 , nxc
               uc(ic,jc,1)   = uf(2 * ic - 1, jf, 1)
               uc(ic,jc,nzc) = uf(2 * ic - 1, jf, 2 * nzc - 1)
            end do
         end do

         if (excttype .eq. 1) call dmlbs(uc,0.0d0,nxc,nyc,nzc)


c-------------------------------------------------------------------
c  Inner boundaries:  Call f2cibfill
c-------------------------------------------------------------------
           call f2cibfill(uf,uc,nxc,nyc,nzc,mask_f,mask_c)


900      continue

         return
      end


c===========================================================
c     2 : 1 linear interpolation of coarse 2--function 'uc' 
c     to fine function 'uf' (prolongation operator).
c===========================================================
      subroutine dmlint(uc,uf,nxc,nyc,nzc)

         implicit       none

         integer        nxc,      nyc,     nzc
         real*8         uc(nxc,nyc,nzc),
     &                  uf(2 * nxc - 1,2 * nyc - 1, 2 * nzc - 1)

         integer        nxf,      nyf,      nzf
         integer        ic,       jc,       if,       jf
         integer        kc,       kf

         nxf = 2 * nxc - 1
         nyf = 2 * nyc - 1
         nzf = 2 * nzc - 1

c   On FG points that correspond to CG points, just copy
         do kc = 1, nzc
            kf = 2 * kc - 1
            do jc = 1 , nyc
               jf = 2 * jc - 1
               do ic = 1 , nxc
                  uf(2 * ic - 1,jf,kf) = uc(ic,jc,kc)
               end do
             end do
          end do

c    Interpolate in the x-direction between CG points
         do kf = 1 , nzf , 2
           do jf = 1 , nyf , 2
             do if = 2 , nxf - 1 , 2
               uf(if,jf,kf) = 0.5d0 * (uf(if+1,jf,kf) + uf(if-1,jf,kf))
             end do
           end do
         end do

c    Interpolate in the y-direction
         do kf = 1 , nzf , 2
           do jf = 2 , nyf - 1 , 2
              do if = 1 , nxf
               uf(if,jf,kf) = 0.5d0 * (uf(if,jf+1,kf) + uf(if,jf-1,kf))
              end do
           end do
         end do

c    Interpolate in the z-direction
         do kf = 2 , nzf-1 , 2
           do jf = 1 , nyf 
              do if = 1 , nxf
               uf(if,jf,kf) = 0.5d0 * (uf(if,jf,kf+1) + uf(if,jf,kf-1))
              end do
           end do
         end do
 
         call gftout(uf,'uf_int',nxf,nyf,nzf)

         return

      end

c===========================================================
c     clcu2: Calculates continuum solution
c===========================================================
      subroutine clcu2(u,x,y,z,nx,ny,nz,
     &                 mask,in_ex,in_ib,in_int,in_esr,in_ob)

         implicit       none

         include       'mgprob.inc'
         include       'maskcodes.inc'

         integer        nx,       ny,        nz
         real*8         u(nx,ny,nz)
         real*8         x(nx),    y(ny),     z(nz)
         integer        mask(nx,ny,nz)
c   Flags to determine whether it's allowed to calculate exact solution
c   there
         logical        in_ex,    in_ib,     in_int,  in_esr,  in_ob

c  Local variables
         integer        i,        j,         k

         do  k = 1 , nz
            do  j = 1 , ny
               do  i = 1 , nx
                  if ( (mask(i,j,k).eq. m_ex .and. in_ex) .or.
     &                 (mask(i,j,k).eq. m_ib .and. in_ib) .or.
     &                 (mask(i,j,k).eq. m_int .and. in_int) .or.
     &                 (mask(i,j,k).eq. m_esr .and. in_esr) .or.
     &                 (mask(i,j,k).eq. m_ob .and. in_ob)) then
                     if (excttype .eq. 2) then
                        u(i,j,k) =   1.0d0 + 2.0d0 * M
     &                         / sqrt(x(i)**2 + y(j)**2 + z(k)**2)
                     else if (excttype .eq. 1) then
                        u(i,j,k) = sin(pi * lx * x(i)) *
     &                             sin(pi * ly * y(j)) *
     &                             sin(pi * lz * z(k))
                     else 
                        write(0,*) 'clcu2: Error, unknown excttype=',
     &                              excttype
                        stop
                     endif
                  endif
               end do
            end do
         end do

         return

      end

c===========================================================
c     Calculates solution error, given continuum solution
c===========================================================
      subroutine calcse(u, se, mask, x, y, z, nx, ny, nz)

         implicit       none
    
         include        'maskcodes.inc'

         integer        nx,       ny,        nz
         real*8         u(nx,ny,nz), se(nx,ny,nz)
         integer        mask(nx,ny,nz)
         real*8         x(nx),    y(ny),     z(nz)

c    Put exact solution into variable "se"
         call clcu2(se,x,y,z,nx,ny,nz,
     &        mask, .true., .true., .true., .true., .true.)

c    Get rid of the infinity in the middle of the mask
         call fillm(se, mask, x, y, z, nx, ny, nz)

c    Subtract u from exact solution and put result in se
         call dmms(se,u,se,nx,ny,nz)

         return
      end


c===========================================================
c     Calculates right hand side corresponding to above 
c     exact solution.
c===========================================================
      subroutine clcrhs2(rhs,x,y,z,nx,ny,nz)

         implicit       none

         include       'mgprob.inc'

         integer        nx,       ny,       nz
         real*8         rhs(nx,ny,nz)
         real*8         x(nx),    y(ny),     z(nz)


         real*8         uijk
         integer        i,        j,         k

         if (excttype .eq. 2) then
            do  k = 1 , nz
              do  j = 1 , ny
                do  i = 1 , nx
                     uijk =   1.0d0 + 2.0d0 * M /
     &                  sqrt(x(i)**2 + y(j)**2 + z(k)**2)

c  SHH: putting if in loop is very slow and should be changed
                    if (eqtype .eq. 1) then
                        rhs(i,j,k) = sigma * uijk**2
                  else if (eqtype .eq. 2) then
                        rhs(i,j,k) = - Ksq * uijk**5 + Asq / uijk**7
                  else
                      write(0,*) 'clcrhs2: Error, unknown eqtype = ',
     &                           eqtype
                      stop
                  endif

                 enddo
              end do
            end do
         else if (excttype .eq. 1) then
            do  k = 1 , nz
               do  j = 1 , ny   
                  do  i = 1 , nx
                     uijk =
     &                     sin(pi * lx * x(i)) *
     &                     sin(pi * ly * y(j)) *
     &                     sin(pi * lz * z(k))
                     rhs(i,j,k) = (-pi**2 * (lx**2 + ly**2 + lz**2) +
     &                     sigma * uijk) * uijk
                  end do
               end do
            end do
         else
           write(0,*) 'clcrhs2: Error, unknown excttype = ',excttype
           stop
         endif

         return

      end




c===========================================================
c     Calculates truncation error estimate
c
c               2h      2h 2h h     2h h  h
c           tau    :=  L  I  u  -  I  L  u
c               h         h        h
c
c           As pointed out in 'mg.inc', this implementation
c           uses three temporary arrays for clarity (:-));
c           production implementation should economize.
c
c  SHH:  Here we DO economize, by using tau as a temporary
c         variable
c===========================================================
      subroutine calctau(q,qsize,m)

      implicit none
 
      include 'mg.inc'

      integer  qsize
      real*8   q(qsize)
    
      integer  m
      
c-----------------------------------------------------------
c             2h           2h h
c           t1     :=     I  u
c                          h
c   We use tau on the coarse grid as a temporary variable:
c             2h        2h  2h
c          tau     :=  L  t1
c                           
c-----------------------------------------------------------
            call dmrshw ( q(u(m)), q(t1(m-1)), nx(m-1), ny(m-1),
     &                    nz(m-1), qm(mask(m)),qm(mask(m-1)),.false.)

            call lop    ( q(tau(m-1)), q(t1(m-1)),qm(mask(m-1)),
     &                    nx(m-1), ny(m-1), nz(m-1), h(m-1))

c-----------------------------------------------------------
c  And use tau on the fine grid as a temporary variable:
c             h            h  h
c          tau     :=     L  u
c                        
c
c             2h        2h   h
c           t1     :=  I  tau
c                       h
c  (SHH: This step is one place where the "copy instead of weighted"
c   condition in the restriction operator is relevant. i.e.
c   if lopu supplies 0's on the inner boundary, then naive weighted 
c   restriction at neighboring points will produce values with lesser
c   magnitudes than perhaps they "should" have)
c-----------------------------------------------------------
            call lop    ( q(tau(m)), q(u(m)), qm(mask(m)),
     &                    nx(m), ny(m), nz(m), h(m))

            call dmrshw ( q(tau(m)), q(t1(m-1)), nx(m-1),ny(m-1),
     &                    nz(m-1), qm(mask(m)), qm(mask(m-1)), .false.)

c-----------------------------------------------------------
c               2h        2h       2h
c           tau    :=  tau    -  t1
c               h
c-----------------------------------------------------------
            call dmms   ( q(tau(m-1)), q(t1(m-1)), q(tau(m-1)),
     &                    nx(m-1), ny(m-1), nz(m-1) )

      return
      end


c===========================================================
c     gsinit: Initializes grid structure.
c===========================================================
      subroutine gsinit(q,qsize,lev,lmin)
 
         implicit       none
 
         include       'mg.inc'
         include       'mgprob.inc'
 
         integer        qalloc 

         integer        qsize
         real*8         q(qsize)
         integer        lev,         lmin
         integer        l,           size
         integer        i,           j,          k
         real*8         xmin,        xmax
         real*8         ymin,        ymax
         real*8         zmin,        zmax

         xmin = xyzmin
         xmax = xyzmax
         ymin = xmin
         ymax = xmax
         zmin = xmin
         zmax = xmax
 
c-----------------------------------------------------------
c        For each level.
c-----------------------------------------------------------
         do l = lmin  , lev
c-----------------------------------------------------------
c           Define mesh dimensions and spacing.
c-----------------------------------------------------------
            if( l .eq. lmin ) then
               nx(lmin)  = 2**lmin + 1
               ny(lmin)  = 2**lmin + 1
               nz(lmin)  = 2**lmin + 1
               h(lmin)   = (xmax - xmin) / (2**lmin)
            else 
               h(l)   = 0.5d0 * h(l-1)
               nx(l)  = 2 * nx(l-1) - 1
               ny(l)  = 2 * ny(l-1) - 1
               nz(l)  = 2 * nz(l-1) - 1
            end if

c-----------------------------------------------------------
c           "Allocate" storage for level-l unknown 'u',  
c           right-hand-side 'rhs', relative local truncation
c           error estimate 'tau', and three temporary 
c           arrays 't1', 't2' and 't3'.
c        SHH: We now only use tau and t1.
c-----------------------------------------------------------
            size    = nx(l) * ny(l) * nz(l)
            write(0,*) 'gsinit: Allocating u, level ',l
            u(l)    = qalloc(q,qsize,size)
            write(0,*) 'gsinit: Allocating rhs, level ',l
            rhs(l)  = qalloc(q,qsize,size)
            write(0,*) 'gsinit: Allocating tau, level ',l
            tau(l)  = qalloc(q,qsize,size)
            write(0,*) 'gsinit: Allocating t1, level ',l
            t1(l)   = qalloc(q,qsize,size)
c-----------------------------------------------------------
c           Set up coordinate arrays
c-----------------------------------------------------------
            x(l) = qalloc(q,qsize,nx(l))
            do i = 0, nx(l)-1
               q(x(l)+i) = xmin + i * h(l) 
            enddo

            y(l) = qalloc(q,qsize,ny(l))
            do i = 0, ny(l)-1
               q(y(l)+i) = ymin + i * h(l) 
            enddo

            z(l) = qalloc(q,qsize,nz(l))
            do i = 0, nz(l)-1
               q(z(l)+i) = zmin + i * h(l) 
            enddo
 
         end do
 
         return
 
      end


c===========================================================
c     "memory allocation" routine.
c     
c     'qalloc' returns the current "pointer" (index) into
c     the main storage array 'q' and increments the pointer
c     by size, effectively "allocating" a block of that 
c     size.  If memory is exhausted, the routine prints 
c     an error message and EXITS.
c===========================================================

      integer function qalloc(q,qsize,size)
         implicit       none
      
         include       'mg.inc'

         integer        qsize
         real*8         q(qsize)
         integer        size 

         logical        first
         data           first / .true. /

         save   
   
         if( first ) then
            qptr  = 1
            first = .false.
         end if
         qalloc = qptr
         if( qptr + size .gt. qsize ) then
            write(0,1000) size, qsize
1000        format(' qalloc: Could not allocate block of ',
     &             'size ',i10,'  Total size: ',i10)
            write(0,*) 'Need this much more space: ',size
            write(0,*) 'Only have this much space: ',qsize-qptr
            stop
         end if
         qptr = qptr + size

         return
      end


c===========================================================
c     Loads "border" of 2-D slices of 3-D array with scalar.
c===========================================================
      subroutine dmlbs(a,s,d1,d2,d3)

         implicit      none

         integer       d1,         d2,       d3
         real*8        a(d1,d2,d3)

         real*8        s

         integer       i1,         i2,      i3

         do i3 = 1, d3
            do i2 = 1 , d2
               a(1, i2,i3) = s
               a(d1,i2,i3) = s
            end do
         end do

         do i3 = 1, d3
            do i1 = 1 , d1
               a(i1,1,i3)  = s
               a(i1,d2,i3) = s
            end do
         end do

         do i2 = 1 , d2
            do i1 = 1 , d1
               a(i1,i2,1)  = s
               a(i1,i2,d3) = s
            end do
         end do 

         return

      end


c===========================================================
c     Load 2-D array with scalar.
c===========================================================
      subroutine dmls(a,s,d1,d2,d3)
 
         implicit       none
 
         integer        d1,       d2,      d3
         real*8         a(d1,d2,d3)
         real*8         s
 
         integer        i,        j,       k
 
         do k = 1 , d3
            do j = 1 , d2
               do i = 1 , d1
                  a(i,j,k) = s
               end do      
            end do
         end do
 
         return
 
      end

c===========================================================
c     multiply 2-D array by scalar.
c===========================================================
      subroutine dmmults(a,s,d1,d2,d3)
 
         implicit       none
 
         integer        d1,       d2,     d3
         real*8         a(d1,d2,d3)
         real*8         s
 
         integer        i,        j,      k
 
         do k = 1 , d3
            do j = 1 , d2
               do i = 1 , d1
                  a(i,j,k) = s * a(i,j,k)
               end do      
            end do
         end do
 
         return
 
      end

c===========================================================
c     matrix-matrix add.
c===========================================================
      subroutine dmma(a1,a2,a3,d1,d2,d3)
 
         implicit       none
 
         integer        d1,         d2,      d3
         real*8         a1(d1,d2,d3),  a2(d1,d2,d3), a3(d1,d2,d3)
 
         integer        i,        j,         k
 
         do k = 1 , d3
            do j = 1 , d2
               do i = 1 , d1
                  a3(i,j,k) = a1(i,j,k) + a2(i,j,k)
               end do
            end do
         end do
 
         return
 
      end

c===========================================================
c     matrix-matrix subtract.
c===========================================================
      subroutine dmms(a1,a2,a3,d1,d2,d3)
 
         implicit       none
 
         integer        d1,            d2,           d3
         real*8         a1(d1,d2,d3),  a2(d1,d2,d3), a3(d1,d2,d3)
 
         integer        i,             j,            k
 
         do k = 1 , d3
            do j = 1 , d2
               do i = 1 , d1
                  a3(i,j,k) = a1(i,j,k) - a2(i,j,k)
               end do
            end do
         end do
 
         return
 
      end

c===========================================================
c     matrix-matrix copy.
c===========================================================
      subroutine dmcopy(a1,a2,d1,d2,d3)

         implicit       none

         integer        d1,           d2,           d3
         real*8         a1(d1,d2,d3), a2(d1,d2,d3)
 
         integer        i,            j,            k
 
         do k = 1 , d3
            do j = 1 , d2
               do i = 1 , d1
                  a2(i,j,k) = a1(i,j,k)
               end do
            end do
         end do
 
         return
 
      end

c===========================================================
c     matrix l2-norm.
c===========================================================
      double precision function dmnrm2(a,d1,d2,d3)

         implicit       none

         integer        d1,       d2,         d3
         real*8         a(d1,d2,d3)

         integer        i,        j,          k

         dmnrm2 = 0.0d0

         do k = 1 , d3
            do j = 1 , d2
               do i = 1 , d1
                 dmnrm2 = dmnrm2 + a(i,j,k) * a(i,j,k)
               end do
            end do
         end do
         dmnrm2 = sqrt(dmnrm2 / (d1 * d2 * d3))

         return

      end

c===========================================================
c     Residual tracing routine.   Generates cheesy "2-d"
c     display.
c===========================================================
      subroutine trace(l,string,resnrm)
         implicit      none
 
         character*(*)  string
         integer        l
         real*8         resnrm

         integer        indent
         parameter    ( indent = 2 )
 
         character*128  spaces 
         data           spaces / ' ' /

         write(0,1000)  spaces(1:1 + (l-1) * indent),
     &                  l, string, resnrm
 1000    format(a,'Level ',i2,' ',a,', sweep ',
     &          t50,1p,e15.6)

         return
 
      end

c===========================================================
c     'gft' output routine.
c
c     This routine uses 'gft_out_bbox': bbox = bounding 
c     box which specifies the limits of the grid function's 
c     domain.
c
c         bbox := ( xmin, xmax, ymin, ymax )
c
c     On the SGIs, specify the libraries 
c
c         -lbbhutil -lmfhdf -ldf -ljpeg -lz
c 
c     (in that precise order) at load-time when using 
c     'gft_out_bbox' or any other routines from the  
c     'bbhutil' library.
c===========================================================
      subroutine gftout(func,str,nx,ny,nz)
         implicit      none

         include      'mgprob.inc'

         integer       nx,ny,nz
         real*8        func(nx,ny,nz)
         character*3   str

         logical       first
         real*8        bbox(6),     time
         integer       shape(3)
         

         data          first / .true. /

         save

         if( .not. gft_trace ) return

         if( first ) then
            time = 0.0d0
            bbox(1) = 0.0d0
            bbox(2) = 1.0d0
            bbox(3) = 0.0d0
            bbox(4) = 1.0d0
            bbox(5) = 0.0d0
            bbox(6) = 1.0d0
            first = .false.
         else
            time = time + 1.0d0
         end if

         shape(1) = nx
         shape(2) = ny
         shape(3) = nz

c         call gft_out_bbox(str,time,shape,3,bbox,func)


         return
 
      end


c=============================================================
c Prints out 3D array, for debugging purposes
c=============================================================
      subroutine printa(u, nx, ny, nz)
      
      implicit none

      integer  nx,        ny,        nz
      real*8   u(nx,ny,nz)

      integer  i,          j,         k

      real*8   tmpmax

      write(0,*) 'Array:--------------'
      do k = 1, nz
         do j = 1, ny
           write(0,1000) (u(i,j,k), i= 1, nx)
         end do
         write(0,*)
      end do
1000  format(1P,20E13.5)

      tmpmax = -1.0d0
      do k = 1, nz
         do j = 1, ny
           do i = 1, nx
             if(abs(u(i,j,k)) .gt. abs(tmpmax)) tmpmax=u(i,j,k)
           enddo
         enddo
      enddo
c      write(0,*) 'absmax = ',tmpmax

      return

      end



c=============================================================
c Prints out 3D integer array, for debugging purposes
c=============================================================
      subroutine printia(u, nx, ny, nz)
         
      implicit none
         
      integer  nx,        ny,        nz
      integer   u(nx,ny,nz)
      
      integer  i,          j,         k
     
      write(0,*) 'Array:--------------'
      do k = 1, nz
         do j = 1, ny
           write(0,1000) (u(i,j,k), i= 1, nx)
         end do
         write(0,*)
      end do
1000  format(1P,20I2)
      return
      end


c=============================================================
c   print a 1-d slice from a 3d array
c=============================================================
      subroutine printslice(u, x, y, z, nx, ny, nz, l, lmax, tostderr)
      implicit none

      integer  nx,        ny,        nz
      real*8   u(nx,ny,nz)
      real*8   x(nx),     y(ny),     z(nz)
      integer  l,         lmax 
      logical  tostderr
       
      integer  i,         j,         k

      k = nz/2 + 1
      j = ny/2 + 1
c  + (ny-1)/8
c      write(0,*) 'printslice: k, j = ',k,j

      if (tostderr) then
         do i = 1, nx 
            write(0,*) x(i), u(i,j,k) / 4**(lmax - l)
         enddo
      else
         do i = 1, nx 
            write(l+7,*) x(i), u(i,j,k) / 4**(lmax - l)
         enddo
      endif

            write(l+7,*) ' '
      return
      end


